@@ -167,8 +167,8 @@ struct CsdMediaKeysManagerPrivate
         /* logind stuff */
         GDBusProxy      *logind_proxy;
         gint             inhibit_keys_fd;
-        GSettings        *session_settings;
-        gboolean         use_logind;
+        GSettings        *desktop_session_settings;
+        GSettings        *cinnamon_session_settings;
 
         /* Multihead stuff */
         GdkScreen       *current_screen;
@@ -186,6 +186,9 @@ struct CsdMediaKeysManagerPrivate
 
         guint            start_idle_id;
 
+        GSettings       *media_key_settings;
+        guint            execute_delay_id;
+
         MprisController *mpris_controller;
 
         /* Ubuntu notifications */
@@ -313,10 +316,76 @@ get_keyring_env (CsdMediaKeysManager *manager)
 	return envp;
 }
 
+static GtkWidget *
+create_dummy_window (GdkScreen *screen)
+{
+  GtkWidget *window;
+
+  window = gtk_window_new (GTK_WINDOW_POPUP);
+
+  gtk_window_move (GTK_WINDOW (window), -100, -100);
+  gtk_window_resize (GTK_WINDOW (window), 10, 10);
+  gtk_widget_show (window);
+
+  return window;
+}
+
+static gboolean
+grab_available (void)
+{
+        GdkDisplay *display;
+        GdkScreen *screen;
+        GdkSeat *seat;
+        GdkGrabStatus res;
+        GtkWidget *dummy_window;
+        gboolean available = FALSE;
+
+        display = gdk_display_get_default ();
+        screen = gdk_display_get_default_screen (display);
+        seat = gdk_display_get_default_seat (display);
+
+        dummy_window = create_dummy_window (screen);
+
+        res = gdk_seat_grab (seat,
+                             gtk_widget_get_window (dummy_window),
+                             GDK_SEAT_CAPABILITY_ALL,
+                             FALSE,
+                             NULL,
+                             NULL,
+                             NULL,
+                             NULL);
+
+        if (res == GDK_GRAB_SUCCESS) {
+                available = TRUE;
+                gdk_seat_ungrab (seat);
+        }
+
+        gtk_widget_destroy (dummy_window);
+
+        return available;
+}
+
+typedef struct {
+        CsdMediaKeysManager *manager;
+        gchar               *cmd;
+        gboolean             need_term;
+        gboolean             need_grab;
+        gint                 grab_try_count;
+} ExecuteData;
+
+#define MAX_GRAB_ATTEMPTS 3
+
 static void
-execute (CsdMediaKeysManager *manager,
-         char                *cmd,
-         gboolean             need_term)
+free_exec_data (ExecuteData *data)
+{
+    g_free (data->cmd);
+    g_slice_free (ExecuteData, data);
+}
+
+static void
+do_execute (CsdMediaKeysManager *manager,
+            const gchar         *cmd,
+            gboolean             need_term)
 {
         gboolean retval;
         char   **argv;
@@ -327,8 +396,9 @@ execute (CsdMediaKeysManager *manager,
 
         retval = FALSE;
 
-        if (need_term)
+        if (need_term) {
                 term = get_term_command (manager);
+        }
 
         if (term) {
                 exec = g_strdup_printf ("%s %s", term, cmd);
@@ -338,9 +408,9 @@ execute (CsdMediaKeysManager *manager,
         }
 
         if (g_shell_parse_argv (exec, &argc, &argv, NULL)) {
-		char   **envp;
+                char   **envp;
 
-		envp = get_keyring_env (manager);
+                envp = get_keyring_env (manager);
 
                 retval = g_spawn_async (g_get_home_dir (),
                                         argv,
@@ -359,9 +429,64 @@ execute (CsdMediaKeysManager *manager,
                 g_warning ("Couldn't execute command: %s: %s", exec, error->message);
                 g_error_free (error);
         }
+
         g_free (exec);
 }
 
+static gboolean
+execute_callback (gpointer data)
+{
+        ExecuteData *exec_data = (ExecuteData *) data;
+        CsdMediaKeysManager *manager = exec_data->manager;
+
+        if (exec_data->need_grab && !grab_available ()) {
+            if (exec_data->grab_try_count < MAX_GRAB_ATTEMPTS) {
+                    exec_data->grab_try_count++;
+
+                    return G_SOURCE_CONTINUE;
+            } else {
+                g_warning ("Unable to grab the keyboard/mouse prior to running: %s", exec_data->cmd);
+
+                free_exec_data (exec_data);
+                manager->priv->execute_delay_id = 0;
+
+                return G_SOURCE_REMOVE;
+            }
+        }
+
+        do_execute (manager, exec_data->cmd, exec_data->need_term);
+
+        free_exec_data (exec_data);
+        manager->priv->execute_delay_id = 0;
+
+        return G_SOURCE_REMOVE;
+}
+
+static void
+execute (CsdMediaKeysManager *manager,
+         gchar               *command,
+         gboolean             need_term,
+         gboolean             need_grab)
+{
+    CsdMediaKeysManagerPrivate *priv = manager->priv;
+    gint delay;
+
+    if (priv->execute_delay_id > 0) {
+            g_source_remove (priv->execute_delay_id);
+    }
+
+    ExecuteData *data = g_slice_new0 (ExecuteData);
+
+    data->manager = manager;
+    data->cmd = g_strdup (command);
+    data->need_term = need_term;
+    data->need_grab = need_grab;
+
+    delay = g_settings_get_uint (priv->media_key_settings, "exec-delay");
+
+    priv->execute_delay_id = g_timeout_add (delay, (GSourceFunc) execute_callback, data);
+}
+
 static void 
 ensure_cancellable (GCancellable **cancellable)
 {
@@ -518,12 +643,28 @@ do_terminal_action (CsdMediaKeysManager *manager)
         term = g_settings_get_string (settings, "exec");
 
         if (term)
-        execute (manager, term, FALSE);
+        execute (manager, term, FALSE, FALSE);
 
         g_free (term);
         g_object_unref (settings);
 }
 
+static void
+do_calculator_action (CsdMediaKeysManager *manager)
+{
+        GSettings *settings;
+        char *calc;
+
+        settings = g_settings_new ("org.cinnamon.desktop.default-applications.calculator");
+        calc = g_settings_get_string (settings, "exec");
+
+        if (calc)
+        execute (manager, calc, FALSE, FALSE);
+
+        g_free (calc);
+        g_object_unref (settings);
+}
+
 static void
 cinnamon_session_shutdown (CsdMediaKeysManager *manager)
 {
@@ -532,7 +673,7 @@ cinnamon_session_shutdown (CsdMediaKeysManager *manager)
 
 	/* Shouldn't happen, but you never know */
 	if (manager->priv->connection == NULL) {
-		execute (manager, "cinnamon-session-quit --logout", FALSE);
+		execute (manager, "cinnamon-session-quit --logout", FALSE, FALSE);
 		return;
 	}
 
@@ -558,7 +699,7 @@ cinnamon_session_shutdown (CsdMediaKeysManager *manager)
 static void
 do_logout_action (CsdMediaKeysManager *manager)
 {
-        execute (manager, "cinnamon-session-quit --logout", FALSE);
+        execute (manager, "cinnamon-session-quit --logout", FALSE, FALSE);
 }
 
 static void
@@ -632,19 +773,13 @@ static void
 do_home_key_action (CsdMediaKeysManager *manager,
 		    gint64               timestamp)
 {
-	GFile *file;
-	GError *error = NULL;
-	char *uri;
+    gchar *path;
 
-	file = g_file_new_for_path (g_get_home_dir ());
-	uri = g_file_get_uri (file);
-	g_object_unref (file);
+    path = g_strdup_printf ("xdg-open %s", g_get_home_dir ());
 
-	if (gtk_show_uri (NULL, uri, timestamp, &error) == FALSE) {
-		g_warning ("Failed to launch '%s': %s", uri, error->message);
-		g_error_free (error);
-	}
-	g_free (uri);
+    execute (manager, path, FALSE, FALSE);
+
+    g_free (path);
 }
 
 static void
@@ -1453,25 +1588,29 @@ do_config_power_action (CsdMediaKeysManager *manager,
                         const gchar *config_key)
 {
         CsdPowerActionType action_type;
-
         action_type = g_settings_get_enum (manager->priv->power_settings,
                                            config_key);
         switch (action_type) {
         case CSD_POWER_ACTION_SUSPEND:
-                csd_power_suspend (manager->priv->use_logind, manager->priv->upower_proxy);
+                ;
+                gboolean hybrid = g_settings_get_boolean (manager->priv->cinnamon_session_settings,
+                                                          "prefer-hybrid-sleep");
+                gboolean suspend_then_hibernate = g_settings_get_boolean (manager->priv->cinnamon_session_settings,
+                                                          "suspend-then-hibernate");
+
+                csd_power_suspend (hybrid, suspend_then_hibernate);
                 break;
         case CSD_POWER_ACTION_INTERACTIVE:
                 cinnamon_session_shutdown (manager);
                 break;
         case CSD_POWER_ACTION_SHUTDOWN:
-                //FIXME: A wee bit cheating here...
-                execute (manager, "dbus-send --dest=org.gnome.SessionManager /org/gnome/SessionManager org.gnome.SessionManager.RequestShutdown", FALSE);
+                csd_power_poweroff ();
                 break;
         case CSD_POWER_ACTION_HIBERNATE:
-                csd_power_hibernate (manager->priv->use_logind, manager->priv->upower_proxy);
+                csd_power_hibernate ();
                 break;
         case CSD_POWER_ACTION_BLANK:
-                execute (manager, "cinnamon-screensaver-command --lock", FALSE);
+                execute (manager, "cinnamon-screensaver-command --lock", FALSE, FALSE);
                 break;
         case CSD_POWER_ACTION_NOTHING:
                 /* these actions cannot be handled by media-keys and
@@ -1681,28 +1820,28 @@ do_action (CsdMediaKeysManager *manager,
                 do_url_action (manager, "mailto", timestamp);
                 break;
         case C_DESKTOP_MEDIA_KEY_SCREENSAVER:
-                execute (manager, "cinnamon-screensaver-command --lock", FALSE);
+                execute (manager, "cinnamon-screensaver-command --lock", FALSE, FALSE);
                 break;
         case C_DESKTOP_MEDIA_KEY_HELP:
                 do_url_action (manager, "ghelp", timestamp);
                 break;
         case C_DESKTOP_MEDIA_KEY_SCREENSHOT:
-                execute (manager, "gnome-screenshot", FALSE);
+                execute (manager, "gnome-screenshot", FALSE, TRUE);
                 break;
         case C_DESKTOP_MEDIA_KEY_WINDOW_SCREENSHOT:
-                execute (manager, "gnome-screenshot --window", FALSE);
+                execute (manager, "gnome-screenshot --window", FALSE, TRUE);
                 break;
         case C_DESKTOP_MEDIA_KEY_AREA_SCREENSHOT:
-                execute (manager, "gnome-screenshot --area", FALSE);
+                execute (manager, "gnome-screenshot --area", FALSE, TRUE);
                 break;
         case C_DESKTOP_MEDIA_KEY_SCREENSHOT_CLIP:
-                execute (manager, "gnome-screenshot --clipboard", FALSE);
+                execute (manager, "gnome-screenshot --clipboard", FALSE, TRUE);
                 break;
         case C_DESKTOP_MEDIA_KEY_WINDOW_SCREENSHOT_CLIP:
-                execute (manager, "gnome-screenshot --window --clipboard", FALSE);
+                execute (manager, "gnome-screenshot --window --clipboard", FALSE, TRUE);
                 break;
         case C_DESKTOP_MEDIA_KEY_AREA_SCREENSHOT_CLIP:
-                execute (manager, "gnome-screenshot --area --clipboard", FALSE);
+                execute (manager, "gnome-screenshot --area --clipboard", FALSE, TRUE);
                 break;
         case C_DESKTOP_MEDIA_KEY_TERMINAL:
                 do_terminal_action (manager);
@@ -1714,15 +1853,7 @@ do_action (CsdMediaKeysManager *manager,
                 do_media_action (manager, timestamp);
                 break;
         case C_DESKTOP_MEDIA_KEY_CALCULATOR:
-                if ((cmd = g_find_program_in_path ("gnome-calculator"))) {
-                execute (manager, "gnome-calculator", FALSE);
-                } else if ((cmd = g_find_program_in_path ("galculator"))) {
-                execute (manager, "galculator", FALSE);
-                } else {
-                execute (manager, "mate-calc", FALSE);
-                }
-
-                g_free (cmd);
+		do_calculator_action (manager);
                 break;
         case C_DESKTOP_MEDIA_KEY_PLAY:
                 return do_multimedia_player_action (manager, NULL, "Play");
@@ -1829,11 +1960,11 @@ start_media_keys_idle_cb (CsdMediaKeysManager *manager)
                                  CA_PROP_APPLICATION_ID, "org.gnome.VolumeControl",
                                  NULL);
 
-        manager->priv->session_settings = g_settings_new("org.cinnamon.desktop.session");
-        manager->priv->use_logind = g_settings_get_boolean (manager->priv->session_settings, "settings-daemon-uses-logind");
-
+        manager->priv->desktop_session_settings = g_settings_new("org.cinnamon.desktop.session");
+        manager->priv->cinnamon_session_settings = g_settings_new("org.cinnamon.SessionManager");
         /* for the power plugin interface code */
         manager->priv->power_settings = g_settings_new (SETTINGS_POWER_DIR);
+        manager->priv->media_key_settings = g_settings_new ("org.cinnamon.settings-daemon.plugins.media-keys");
 
         /* Logic from http://git.gnome.org/browse/gnome-shell/tree/js/ui/status/accessibility.js#n163 */
         manager->priv->interface_settings = g_settings_new (SETTINGS_INTERFACE_DIR);
@@ -1951,6 +2082,26 @@ csd_media_keys_manager_stop (CsdMediaKeysManager *manager)
                 priv->power_settings = NULL;
         }
 
+        if (priv->desktop_session_settings) {
+                g_object_unref (priv->desktop_session_settings);
+                priv->desktop_session_settings = NULL;
+        }
+
+        if (priv->cinnamon_session_settings) {
+                g_object_unref (priv->cinnamon_session_settings);
+                priv->cinnamon_session_settings = NULL;
+        }
+
+        if (priv->interface_settings) {
+                g_object_unref (priv->interface_settings);
+                priv->interface_settings = NULL;
+        }
+
+        if (priv->media_key_settings) {
+                g_object_unref (priv->media_key_settings);
+                priv->media_key_settings = NULL;
+        }
+
         if (priv->power_screen_proxy) {
                 g_object_unref (priv->power_screen_proxy);
                 priv->power_screen_proxy = NULL;
